"""
Helpers for testing your regret.
"""
from __future__ import annotations

from contextlib import AbstractContextManager, contextmanager
from typing import Any, Iterator

from attrs import field, frozen

from regret.emitted import Deprecation
from regret.typing import Deprecatable


class ExpectedDifferentDeprecations(AssertionError):
    """
    Different deprecation(s) were seen than the ones which were expected.
    """


@frozen
class Recorder:
    """
    Recorders keep track of deprecations as they are emitted.

    They provide helper methods for asserting about the deprecations
    afterwards.
    """

    _saw: list[Deprecatable] = field(factory=list, alias="saw")

    def emit(self, deprecation: Deprecatable, extra_stacklevel: int) -> None:
        """
        "Emit" a deprecation by simply storing it.

        An emitter suitable for passing to `regret.Deprecator` instances.
        """
        self._saw.append(deprecation)

    def expect(self, **kwargs: Any) -> AbstractContextManager[None]:
        """
        Expect a given set of deprecations to be emitted.
        """
        return self.expect_deprecations(Deprecation(**kwargs))

    @contextmanager
    def expect_deprecations(
        self,
        *deprecations: Deprecation,
    ) -> Iterator[None]:
        """
        Expect a given set of deprecations to be emitted.
        """
        expected = self._saw + list(deprecations)
        yield
        if self._saw != expected:
            raise ExpectedDifferentDeprecations((self._saw, expected))

    def expect_clean(self) -> AbstractContextManager[None]:
        """
        Expect no deprecations to be emitted.
        """
        return self.expect_deprecations()



import importlib.metadata
import re

project = "regret"
author = "Julian Berman"
copyright = "2019, " + author

release = importlib.metadata.version("regret")
version = release.partition("-")[0]

language = "en"
default_role = "any"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx_copybutton",
    "sphinxcontrib.spelling",
    "sphinxext.opengraph",
]

pygments_style = "lovelace"
pygments_dark_style = "one-dark"

html_theme = "furo"


def entire_domain(host):
    return r"http.?://" + re.escape(host) + r"($|/.*)"


linkcheck_ignore = [
    entire_domain("img.shields.io"),
    "https://github.com/Julian/regret/actions",
    "https://github.com/Julian/regret/workflows/CI/badge.svg",
]


autodoc_default_options = {
    "members": True,
    "member-order": "bysource",
}

# -- autosectionlabel ----------------------------------

autosectionlabel_prefix_document = True

# -- intersphinx ---------------------------------------

intersphinx_mapping = {
    "attrs": ("https://attrs.org/en/stable/", None),
    "cpython-devguide": ("https://devguide.python.org/", None),
    "jsonschema": (
        "https://python-jsonschema.readthedocs.io/en/latest/",
        None,
    ),
    "packaging": ("https://packaging.python.org/en/latest/", None),
    "python": ("https://docs.python.org/3/", None),
    "requests": ("https://requests.readthedocs.io/en/latest/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}



"""
Typing related helpers for regret.
"""
from __future__ import annotations

from datetime import date
from typing import Any, Callable, Protocol


class Deprecatable(Protocol):
    """
    A single kind of deprecatable behavior.
    """

    def message(self, name_of: name_of) -> str:
        """
        Return a message summarizing this deprecation.
        """
        ...


class Emitter(Protocol):
    """
    A callable which reports when deprecated things are used.
    """

    def __call__(
        self,
        extra_stacklevel: int = ...,
        **kwargs: Any,
    ) -> None:
        """
        Somehow emit that something deprecated has been used.
        """
        ...


name_of = Callable[[Any], str]


class new_docstring(Protocol):
    """
    A callable which transforms docstrings to include deprecation info.
    """

    def __call__(
        self,
        object: Any,
        name_of: name_of,
        replacement: Any | None,
        removal_date: date | None,
        version: str,
    ) -> str:
        """
        Transform the docstring, producing a new one.
        """
        ...



from __future__ import annotations

from datetime import date
from textwrap import dedent
from typing import Any, Callable


def doc_with_deprecated_directive(
    object: Any,
    replacement: Any | None,
    removal_date: date | None,
    name_of: Callable[..., str],
    version: str,
):
    """
    Add a `deprecated` directive to the provided object's docstring.

    Suitable for use with `regret.Deprecator`.
    """
    parts = [
        dedent(object.__doc__),
        f"\n.. deprecated:: {version}\n",
    ]
    if replacement is not None:
        parts.append(
            f"\n    Please use `{name_of(replacement)}` instead.\n",
        )
    if removal_date is not None:
        parts.append(
            f"\n    It will be removed on or after {removal_date}.\n",
        )
    return "".join(parts)



"""
Objects emitted whilst a deprecated object is being used.
"""
from __future__ import annotations

from datetime import date
from typing import Any, Callable as _Callable
import inspect

from attrs import field, frozen

from regret.typing import Deprecatable, name_of


def _qualname(obj: Any) -> str:
    """
    Return the (non-fully-)qualified name of the given object.
    """
    return obj.__qualname__


@frozen
class Deprecation:
    """
    A single emitted deprecation.
    """

    _kind: Deprecatable = field(alias="kind")
    _name_of: name_of = field(
        default=_qualname,
        repr=False,
        alias="name_of",
    )
    _replacement: Any = field(default=None, repr=False, alias="replacement")
    _removal_date: date | None = field(
        default=None,
        repr=False,
        alias="removal_date",
    )
    _addendum: str | None = field(default=None, repr=False, alias="addendum")

    def message(self) -> str:
        """
        Express this deprecation as a comprehensible message.
        """
        parts: list[str] = [self._kind.message(name_of=self._name_of)]
        if self._removal_date is not None:
            parts.append(
                f"It will be removed on or after {self._removal_date}.",
            )
        if self._replacement is not None:
            parts.append(
                f"Please use {self._name_of(self._replacement)} instead.",
            )
        if self._addendum is not None:
            parts.append(self._addendum)
        return " ".join(parts)


# --* Representations of deprecated things *--


@frozen
class Callable:
    """
    A parameter for a particular callable.
    """

    _object: Any = field(alias="object")

    def message(self, name_of: name_of) -> str:
        """
        Express this deprecation as a comprehensible message.
        """
        return f"{name_of(self._object)} is deprecated."


@frozen
class Inheritance:
    """
    The subclassing of a given parent type.
    """

    _type: type = field(alias="type")

    def message(self, name_of: name_of) -> str:
        """
        Express this deprecation as a comprehensible message.
        """
        return f"Subclassing from {name_of(self._type)} is deprecated."


@frozen
class Parameter:
    """
    A parameter for a particular callable which should no longer be used.
    """

    _callable: _Callable[..., Any] = field(alias="callable")
    _parameter: inspect.Parameter = field(alias="parameter")

    def message(self, name_of: name_of) -> str:
        """
        Express this deprecation as a comprehensible message.
        """
        return f"The {self._parameter.name!r} parameter is deprecated."


@frozen
class OptionalParameter:
    """
    A parameter for a particular callable which will become mandatory.
    """

    _callable: _Callable[..., Any] = field(alias="callable")
    _parameter: inspect.Parameter = field(alias="parameter")
    _default: Any = field(alias="default")

    def message(self, name_of: name_of) -> str:
        """
        Express this deprecation as a comprehensible message.
        """
        return (
            f"Calling {name_of(self._callable)} without providing the "
            f"{self._parameter.name!r} parameter is deprecated. Using "
            f"{self._default!r} as a default."
        )



"""
Integration with the standard library's `warnings` module.
"""
import warnings

from regret.emitted import Deprecation

_STACKLEVELS_UNTIL_EMIT_IS_CALLED = 4


def emit(deprecation: Deprecation, extra_stacklevel: int):
    warnings.warn(
        deprecation.message(),
        DeprecationWarning,
        stacklevel=_STACKLEVELS_UNTIL_EMIT_IS_CALLED + extra_stacklevel,
    )



"""
You made a thing, but now you wish it'd go away... Deprecations, a love story.
"""
from regret._api import _DEPRECATOR  # type: ignore[reportPrivateUsage]
from regret._api import Deprecator

callable = _DEPRECATOR.callable
inheritance = _DEPRECATOR.inheritance
parameter = _DEPRECATOR.parameter
optional_parameter = _DEPRECATOR.optional_parameter

__all__ = [
    "Deprecator",
    "callable",
    "inheritance",
    "optional_parameter",
    "parameter",
]



"""
Signature helpers for deprecated (or partially-deprecated) callables.
"""
from __future__ import annotations

from typing import Any, Callable, Iterable
import inspect

from attrs import evolve, field, frozen


class AlreadyDeprecated(Exception):
    """
    An attempt was made to deprecate an already-deprecated parameter.
    """


class NoSuchParameter(Exception):
    """
    An attempt was made to deprecate a parameter that does not exist.
    """


@frozen
class SignatureWithRegret:
    """
    An `inspect.Signature`, along with its regretted subparts.
    """

    _signature: inspect.Signature = field(alias="signature")
    _deprecated: list[str] = field(factory=list, alias="deprecated")
    _defaults_for_optional_parameters: dict[str, Any] = field(
        factory=dict,
        alias="defaults_for_optional_parameters",
    )
    kwargs_parameter_name: str | None = field(init=False)
    _order: dict[str | None, int] = field(init=False)

    def __attrs_post_init__(self) -> None:
        object.__setattr__(
            self,
            "kwargs_parameter_name",
            next(
                (
                    name
                    for name, parameter in reversed(
                        self._signature.parameters.items(),
                    )
                    if parameter.kind == inspect.Parameter.VAR_KEYWORD
                ),
                None,
            ),
        )

        object.__setattr__(
            self,
            "_order",
            {
                parameter: index
                for index, parameter in enumerate(self._signature.parameters)
            },
        )

    @classmethod
    def for_callable(
        cls,
        callable: Callable[..., Any],
        **kwargs: Any,
    ) -> SignatureWithRegret:
        return cls(
            signature=inspect.signature(callable, follow_wrapped=False),
            **kwargs,
        )

    def would_accept(self, name: str):
        """
        Does this signature know about a parameter with the given name?
        """  # noqa: D401
        return (
            name in self._signature.parameters
            or self.kwargs_parameter_name is not None
        )

    def with_parameter(self, name: str, **kwargs: Any) -> SignatureWithRegret:
        """
        Evolve this signature to add a deprecated parameter.
        """
        if not self.would_accept(name):
            raise NoSuchParameter(name)
        elif name in self._deprecated:
            raise AlreadyDeprecated(name)

        deprecated = sorted(
            self._deprecated + [name],
            key=lambda each: (
                self._order.get(
                    each,
                    self._order.get(self.kwargs_parameter_name, -1),
                ),
                each,
            ),
        )
        return evolve(self, deprecated=deprecated, **kwargs)

    def with_optional_parameter(self, name: str, default: Any):
        """
        Evolve this signature to add a deprecated optional parameter.
        """
        return self.with_parameter(
            name,
            defaults_for_optional_parameters={
                name: default,
                **self._defaults_for_optional_parameters,
            },
        )

    def misused(
        self,
        bound_arguments: inspect.BoundArguments,
        callable: Callable[..., Any],
    ) -> Iterable[tuple[inspect.Parameter, bool]]:
        """
        Collect the arguments (optional or required) that are misused.
        """
        arguments = bound_arguments.arguments
        kwargs = bound_arguments.arguments.get(self.kwargs_parameter_name, ())  # type: ignore[reportGeneralTypeIssues]  # noqa: E501

        for name in self._deprecated:
            is_optional = name in self._defaults_for_optional_parameters
            if is_optional:
                if name in arguments or name in kwargs:
                    continue

                parameter = self._signature.parameters.get(name)
                if parameter is None:
                    if self.kwargs_parameter_name is None:
                        continue
                    else:
                        parameter = inspect.Parameter(
                            name=name,
                            kind=inspect.Parameter.KEYWORD_ONLY,
                        )
                yield parameter, is_optional
            else:
                if name in arguments:
                    yield self._signature.parameters[name], is_optional
                elif name in kwargs:
                    yield inspect.Parameter(
                        name=name,
                        kind=inspect.Parameter.KEYWORD_ONLY,
                    ), is_optional

    def bind(self, *args: Any, **kwargs: Any) -> inspect.BoundArguments:
        return self._signature.bind_partial(*args, **kwargs)

    def set_default(
        self,
        bound_arguments: inspect.BoundArguments,
        parameter: inspect.Parameter,
    ) -> Any:
        """
        Set the default for the given parameter within some bound arguments.
        """
        name = parameter.name
        default = self._defaults_for_optional_parameters[name]
        if name in self._signature.parameters:
            bound_arguments.arguments[name] = default
        elif self.kwargs_parameter_name is None:
            raise TypeError(
                f"No parameter {name} and no kwargs accepted.",
            )
        else:
            kwargs: dict[str, Any] = bound_arguments.arguments.setdefault(
                self.kwargs_parameter_name,
                {},
            )
            kwargs[name] = default
        return default
